@@ -7,6 +7,7 @@ from remerkleable.core import BackedView, FixedByteLengthViewHelper, \
 from remerkleable.tree import Node, PairNode, zero_node, Gindex, to_gindex, Link, RootNode, NavigationError,\
     Root, subtree_fill_to_contents, subtree_fill_to_length, get_depth
 from remerkleable.basic import boolean, uint256
+from remerkleable.readonly_iters import BitfieldIter
 
 V = TypeVar('V', bound=View)
 
@@ -56,9 +57,6 @@ class BitsView(BackedView, ColSequence):
     def __len__(self):
         return self.length()
 
-    def __iter__(self):
-        return iter(self.get(i) for i in range(self.length()))
-
     def __getitem__(self, k):
         length = self.length()
         if isinstance(k, slice):
@@ -125,6 +123,9 @@ class Bitlist(BitsView):
 
         return SpecialBitlistView
 
+    def __iter__(self):
+        return BitfieldIter(self.get_backing().get_left(), self.contents_depth(), self.length())
+
     @classmethod
     def contents_depth(cls) -> int:  # depth excluding the length mix-in
         return get_depth((cls.limit() + 255) // 256)
@@ -347,6 +348,9 @@ class Bitvector(BitsView, FixedByteLengthViewHelper):
 
         return SpecialBitvectorView
 
+    def __iter__(self):
+        return BitfieldIter(self.get_backing(), self.__class__.tree_depth(), self.__class__.vector_length())
+
     @classmethod
     def tree_depth(cls) -> int:
         return get_depth((cls.vector_length() + 255) // 256)
