@@ -1,20 +1,23 @@
 # flake8:noqa F401  Ignore unused imports. Tests are a work in progress.
 import pytest
 
+from random import Random
+
 from remerkleable.core import View, TypeDef, BasicView
 from remerkleable.complex import Container, Vector, List
-from remerkleable.basic import boolean, bit, uint, byte, uint8, uint16, uint32, uint64, uint128, uint256
+from remerkleable.basic import boolean, bit, uint, byte, uint8, uint16, uint32, uint64, uint128, uint256,\
+    OperationNotSupported
 from remerkleable.bitfields import Bitvector, Bitlist
 from remerkleable.byte_arrays import ByteVector, Bytes1, Bytes4, Bytes8, Bytes32, Bytes48, Bytes96
 from remerkleable.core import BasicView, View, TypeDef
 from remerkleable.tree import get_depth
 
 
-def expect_value_error(fn, msg):
+def expect_op_error(fn, msg):
     try:
         fn()
         raise AssertionError(msg)
-    except ValueError:
+    except (ValueError, OperationNotSupported, AttributeError) as e:
         pass
 
 
@@ -62,13 +65,40 @@ def test_basic_value_bounds():
         # this should work
         assert k(v - 1) == v - 1
         # but we do not allow overflows
-        expect_value_error(lambda: k(v), f"no overflows allowed: type: {k}, value: {v}")
+        expect_op_error(lambda: k(v), f"no overflows allowed: type: {k}, value: {v}")
 
     for k, _ in max.items():
         # this should work
         assert k(0) == 0
         # but we do not allow underflows
-        expect_value_error(lambda: k(-1), "no underflows allowed")
+        expect_op_error(lambda: k(-1), "no underflows allowed")
+
+    for k, v in max.items():
+        if v == 2:
+            continue  # skip bool/bit
+        half = v // 2
+        # this should work
+        assert k(half) + k(half-1) == v - 1
+        # but 2x half == max, so this should fail
+        expect_op_error(lambda: k(half) + k(half), f"no __add__ overflows allowed: type: {k}, value: {half}")
+        # this should work
+        assert k(half) - k(half) == 0
+        assert k(half-1) - k(half-1) == 0
+        # but underflow should not
+        expect_op_error(lambda: k(half - 1) - k(half), f"no __sub__ underflows allowed: type: {k}, value: {half}")
+
+    for k, v in max.items():
+        if v == 2:
+            continue  # skip bool/bit
+        assert k(v-1) * k(1) == v-1
+        assert k(v // 3) * 2 == (v // 3) * 2
+        assert k((v // 2) - 1) * 2 == ((v // 2) - 1) * 2
+        # but 2x max-1 should fail
+        expect_op_error(lambda: k(v - 1) * 2, f"no __mul__ overflows allowed: type: {k}")
+        # and 2x half too
+        expect_op_error(lambda: k(v // 2) * 2, f"no __mul__ overflows allowed: type: {k}")
+        expect_op_error(lambda: k(v // 2) // 0.5, f"no __floordiv__ with float overflows allowed: type: {k}")
+        expect_op_error(lambda: k(v - 1) / 2.0, f"no __truediv__ allowed: type: {k}")
 
 
 def test_container():
@@ -108,6 +138,13 @@ def test_container():
     assert v.__class__.element_cls() == uint8
     assert v.__class__.limit() == 1024
 
+    field_values = list(y)
+    assert field_values == [y.a, y.b]
+
+    f_a, f_b = y
+    assert f_a == y.a
+    assert f_b == y.b
+
     y.a = 42
     try:
         y.a = 256  # out of bounds
@@ -127,6 +164,23 @@ def test_container():
     except AttributeError:
         pass
 
+    try:
+        Foo(wrong_field_name=100)
+        assert False
+    except AttributeError:
+        pass
+
+
+def test_container_unpack():
+    class Foo(Container):
+        a: uint64
+        b: uint8
+        c: Vector[uint16, 123]
+
+    foo = Foo(b=42)
+    a, b, c = foo
+    assert b == 42
+
 
 def test_list():
     typ = List[uint64, 128]
@@ -213,11 +267,11 @@ def test_bytesn_subclass():
 def test_uint_math():
     assert uint8(0) + uint8(uint32(16)) == uint8(16)  # allow explicit casting to make invalid addition valid
 
-    expect_value_error(lambda: uint8(0) - uint8(1), "no underflows allowed")
-    expect_value_error(lambda: uint8(1) + uint8(255), "no overflows allowed")
-    expect_value_error(lambda: uint8(0) + 256, "no overflows allowed")
-    expect_value_error(lambda: uint8(42) + uint32(123), "no mixed types")
-    expect_value_error(lambda: uint32(42) + uint8(123), "no mixed types")
+    expect_op_error(lambda: uint8(0) - uint8(1), "no underflows allowed")
+    expect_op_error(lambda: uint8(1) + uint8(255), "no overflows allowed")
+    expect_op_error(lambda: uint8(0) + 256, "no overflows allowed")
+    expect_op_error(lambda: uint8(42) + uint32(123), "no mixed types")
+    expect_op_error(lambda: uint32(42) + uint8(123), "no mixed types")
 
     assert type(uint32(1234) + 56) == uint32
 
@@ -314,6 +368,17 @@ def test_bitvector():
             assert bool(b[i]) == False, "unset %d / %d" % (i, size)
 
 
+def test_bitvector_iter():
+    rng = Random(123)
+    for size in [1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 16, 17, 31, 32, 33, 63, 64, 65, 511, 512, 513, 1023, 1024, 1025]:
+        # get a somewhat random bitvector
+        bools = list(rng.randint(0, 1) == 1 for i in range(size))
+        b = Bitvector[size](*bools)
+        # Test iterator
+        for i, bit in enumerate(b):
+            assert bool(bit) == bools[i]
+
+
 def test_bitlist():
     for size in [1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 16, 17, 31, 32, 33, 63, 64, 65, 511, 512, 513, 1023, 1024, 1025]:
         for i in range(size):
@@ -325,3 +390,15 @@ def test_bitlist():
             b = Bitlist[size](True for i in range(size))
             b[i] = False
             assert bool(b[i]) == False, "unset %d / %d" % (i, size)
+
+
+def test_bitlist_iter():
+    rng = Random(123)
+    for limit in [1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 16, 17, 31, 32, 33, 63, 64, 65, 511, 512, 513, 1023, 1024, 1025]:
+        for length in [0, 1, limit // 2, limit]:
+            # get a somewhat random bitvector
+            bools = list(rng.randint(0, 1) == 1 for i in range(length))
+            b = Bitlist[limit](*bools)
+            # Test iterator
+            for i, bit in enumerate(b):
+                assert bool(bit) == bools[i]
